/*!
 * module STM32CubeMonitor
 * Copyright(c) 2021 STMicroelectronics
 */
// Modules to control application life and create native browser window
const { app, BrowserWindow, dialog, Menu, globalShortcut, shell, ipcMain } = require('electron');
const os = require('os');
// this should be placed at top of main.js to handle squirrel setup events quickly
if (handleSquirrelEvent()) { return; }


const serviceRed = require('./service-red');
const url = require('url');
const path = require('path');
const uiPath = (serviceRed.settings.ui || {}).path || "ui";

// Keep a global reference of the window object, if you don't, the window will
// be closed automatically when the JavaScript object is garbage collected.
let mainWindow, splash, previousBounds;
let isMaximized = false, isFullScreen = false;

// set the applicaton mutex to prevent uninstaller to run (windows only)
var mutex;
if (os.platform() === 'win32') {
  try {
    const windows_mutex = require('windows-mutex');
    mutex = new windows_mutex.Mutex("stm32cubemonitorlock");
  } catch (error) { }
}

process.env.NO_UPDATE_NOTIFIER = 1;

// configure the native windows menu 
if (process.platform !== 'darwin') {
// Remove menu in all windows
Menu.setApplicationMenu(null);
} else {
// For MacOS, keep the edit item in the menu to handle the copy paste shortcuts
// BZ84740

  var menuTemplate = [{
    
    label: "Menu",
    submenu: [
        { label: "Undo", accelerator: "CmdOrCtrl+Z", selector: "undo:" },
        { type: "separator" },
        { label: "Cut", accelerator: "CmdOrCtrl+X", selector: "cut:" },
        { label: "Copy", accelerator: "CmdOrCtrl+C", selector: "copy:" },
        { label: "Paste", accelerator: "CmdOrCtrl+V", selector: "paste:" },
        { label: "Select All", accelerator: "CmdOrCtrl+A", selector: "selectAll:" },
        { type: "separator" },
        { label: "Quit", accelerator: "Command+Q", selector: "quit:"}
    ]}
];
Menu.setApplicationMenu(Menu.buildFromTemplate(menuTemplate));
}

function createWindow() {
  let config = {
    width: 1024,
    height: 720,
    minWidth: 1024 / 2, // enable split screen on macos
    minHeight: 720 / 2, // enable split screen on macos
    backgroundColor: '#2e2c29',
    webPreferences: {
      nodeIntegration: false,
      contextIsolation: true,
      nativeWindowOpen: true
    },
    frame: true,
    icon : path.join(__dirname, 'default/ressources/CubeMonitor.ico')
  }


  // Create the browser window.
  mainWindow = new BrowserWindow(config);

  // Opening node-red editor
  const editor_url = "http://localhost:" + serviceRed.settings.uiPort + "/";
  mainWindow.loadURL(editor_url);


  mainWindow.webContents.on('will-prevent-unload', (event) => {
    let choice = dialog.showMessageBoxSync(mainWindow, {
      type: 'warning',
      buttons: ['OK', 'Cancel'],
      message: 'Do you really want to quit ?',
      detail: 'Your current flow is not deployed, changes will be lost.',
      defaultId: 0,
      cancelId: 1
    });

    const leave = (choice === 0)
    if (leave) {
      event.preventDefault();
    }
  })

  mainWindow.webContents.setWindowOpenHandler(({ url }) => {

    // Open wiki & ST community in the browser
    if (!url.startsWith("http://localhost:" + serviceRed.settings.uiPort + "/" + uiPath)) {
      shell.openExternal(url);
      return { action: "deny" };
    }

    let openedWindows = BrowserWindow.getAllWindows();
    let dashboard = openedWindows.find(function (elem) {
      return mainWindow.id !== elem.id;
    });
    if (dashboard) {
      dashboard.show();
      dashboard.focus();
      dashboard.flashFrame(true);
      // do not create new windows
      return { action: "deny" };
    }

    const pos = mainWindow.getPosition();

    dashboard = new BrowserWindow({
      minWidth: 1024 / 4, // enable split screen on macos
      minHeight: 720 / 4, // enable split screen on macos
      show: false,
      webPreferences: {
        nodeIntegration: false,
        contextIsolation: true,
        nativeWindowOpen: true
      }
    });
    dashboard.loadURL(url);

    // if dashboard already opened & closed previously
    // then restore the size and position

    if (previousBounds) {
      dashboard.setBounds(previousBounds);
    } else {
      dashboard.setBounds({ width: 1024, height: 720, x: pos[0] + 30, y: pos[1] + 30 });
    }
   
    if (isMaximized) {
      dashboard.maximize();
    }

    dashboard.setFullScreen(isFullScreen);

    dashboard.on("maximize", () => {
      isMaximized = true;
    });
    dashboard.on("unmaximize", () => {
      isMaximized = false;
    });

    dashboard.on("close", () => {
      if (dashboard.isNormal()) {
        isMaximized = isFullScreen = false;
        previousBounds = dashboard.getBounds();
      }
    });

    dashboard.on("will-resize", () => {
      previousBounds = dashboard.getBounds();
    });

    dashboard.on("enter-full-screen", () => {
      isFullScreen = true;
    });
    dashboard.on("leave-full-screen", () => {
      isFullScreen = false;
    });

    if (os.platform() === "win32") {
      dashboard.on("will-move", (e, n) => {
        previousBounds = dashboard.getBounds();
      });
    } else {
      dashboard.on("move", () => {
        previousBounds = dashboard.getBounds();
      });
    }

    //enable the dev tool for debug
    //dashboard.webContents.openDevTools();

    dashboard.on("ready-to-show", () => {
      dashboard.show();
    });

    return { action: "deny" };
  });

  // Open the DevTools.
  // mainWindow.webContents.openDevTools();

  // Emitted when the window is closed.
  mainWindow.on('closed', function () {
    // Dereference the window object, usually you would store windows
    // in an array if your app supports multi windows, this is the time
    // when you should delete the corresponding element.
    BrowserWindow.getAllWindows().forEach(w => w.close());
    mainWindow = null
  });

  mainWindow.once('ready-to-show', () => {
    mainWindow.show();

  });

  mainWindow.webContents.on("did-finish-load", () => {
    mainWindow.show();
    setTimeout(() => {
      splash.destroy();
    }, 300);
  });

}


// This method will be called when Electron has finished
// initialization and is ready to create browser windows.
// Some APIs can only be used after this event occurs.
app.once('ready', () => {
  splash = new BrowserWindow({
    width: 647,
    height: 425,
    resizable: false,
    movable: false,
    useContentSize: true,
    backgroundColor: '#CFECFA',
    frame: false,
    alwaysOnTop: true,
    webPreferences: {
      contextIsolation: true,
      nativeWindowOpen: true,
    },
    icon : path.join(__dirname, 'default/ressources/CubeMonitor.ico')
  });


  splash.loadURL(url.format({
    pathname: path.join(__dirname, 'splash.html'),
    protocol: 'file:',
    slashes: true
  }))

  splash.once('ready-to-show', () => {
    splash.show();
  })

  serviceRed.start().then(() => {
    createWindow();
  }).catch(function (err) {
    if (err.errno==="EADDRINUSE") {
    const choice = dialog.showMessageBoxSync(splash, {
      type: 'error',
      buttons: ['OK'],
      message: 'Application server port already in use',
      detail: 'The application server is not able to start due to current port ' + serviceRed.settings.uiPort + ' already in use.\nPlease stop the server using this port before starting this application.',
    })
    app.quit();
  } else {
    const choice = dialog.showMessageBoxSync(splash, {
      type: 'error',
      buttons: ['OK'],
      message: 'Application error',
      detail: 'The application is not able to start due to error :' + err+'.',
    })
    app.quit();
  }
  });
});

// Quit when all windows are closed.
app.on('window-all-closed', function () {
  // check mutex state (keep the mutex in a closure)
  if (os.platform() === 'win32') {
    try {
      mutex.isActive();
    } catch (error) { }
  }

  app.quit()
});

app.on('activate', function () {
  // On OS X it's common to re-create a window in the app when the
  // dock icon is clicked and there are no other windows open.
  if (mainWindow === null) {
    createWindow()
  }
});

app.on('will-quit', () => {
  // Unregister all shortcuts.
  globalShortcut.unregisterAll();
})

app.on('login', (event, webContents, details, authInfo, callback) => {

  //Create the login popup in secured mode only
  event.newGuest = new BrowserWindow({
    width: 630, height: 300,
    parent: mainWindow,
    show: false,
    modal: true,
    alwaysOnTop: true,
    resizable: false,
    title: "Authentication",
    frame: false,
    autoHideMenuBar: true,
    icon: './default/ressources/CubeMonitor.ico',
    webPreferences: {
      nodeIntegration: true,
      sandbox: false,
      preload: path.join(__dirname, 'preload.js')
    }
  });

  // Load and display the login popup
  event.newGuest.loadFile(path.join(__dirname, "popup.html"));
  event.newGuest.once('ready-to-show', () => { event.newGuest.show() });

  //Prevent the default behavior
  if (!event.defaultPrevented) {
    event.preventDefault();
  } else {
    return;
  }

  //Asynchronous function creating a new promise
  function asyncFunction(username, password) {
    return new Promise((resolve, reject) => {
      const credentials = { username, password };
      if (credentials) {
        resolve(credentials);
      } else {
        reject(error);
      }
    });
  }

  //On 'login' click, Get credentials from the renderer process and callback 
  ipcMain.once('credentials-submission', (event, username, password) => {

    let loginPromise = asyncFunction(username, password);

    loginPromise
      .then(credentials => {
        callback(credentials.username, credentials.password);
      })
      .catch(error => {
        console.log(error, "-------Credentials Promise error-------");
      });

    ipcMain.removeAllListeners('cancel');
  });

  //On 'cancel' click, get void credentials from the renderer process and callback 
  ipcMain.once('cancel', (event, username, password) => {

    let  cancelPromise = asyncFunction(username, password);

    cancelPromise
      .then(() => {
        callback();
      })
      .catch(error => {
        console.log(error, "-------Cancel Promise error-------");
      });

    ipcMain.removeAllListeners('credentials-submission');
  });
});


// In this file you can include the rest of your app's specific main process
// code. You can also put them in separate files and require them here.

///////////////////////////////////////////////////////
// All this Squirrel stuff is for the Windows installer
function handleSquirrelEvent() {
  if (process.argv.length === 1) {
    return false;
  }

  const ChildProcess = require('child_process');
  const path = require('path');

  const appFolder = path.resolve(process.execPath, '..');
  const rootAtomFolder = path.resolve(appFolder, '..');
  const updateDotExe = path.resolve(path.join(rootAtomFolder, 'Update.exe'));
  const exeName = path.basename(process.execPath);

  const spawn = function (command, args) {
    let spawnedProcess, error;

    try {
      spawnedProcess = ChildProcess.spawn(command, args, { detached: true });
    } catch (error) { }

    return spawnedProcess;
  };

  const spawnUpdate = function (args) {
    return spawn(updateDotExe, args);
  };

  const squirrelEvent = process.argv[1];
  switch (squirrelEvent) {
    case '--squirrel-install':
    case '--squirrel-updated':
      // Optionally do things such as:
      // - Add your .exe to the PATH
      // - Write to the registry for things like file associations and
      //   explorer context menus

      // Install desktop and start menu shortcuts
      spawnUpdate(['--createShortcut', exeName]);

      setTimeout(app.quit, 1000);
      return true;

    case '--squirrel-uninstall':
      // Undo anything you did in the --squirrel-install and
      // --squirrel-updated handlers

      // Remove desktop and start menu shortcuts
      spawnUpdate(['--removeShortcut', exeName]);

      setTimeout(app.quit, 1000);
      return true;

    case '--squirrel-obsolete':
      // This is called on the outgoing version of your app before
      // we update to the new version - it's the opposite of
      // --squirrel-updated

      app.quit();
      return true;
  }
};
