"use strict";
/*!
 * module acquisitionservices
 * Copyright(c) 2019 STMicroelectronics
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AcquisitionServices = void 0;
const stm32cubemonitor_logger_1 = require("@stm32/stm32cubemonitor-logger");
const logger = (0, stm32cubemonitor_logger_1.getLogger)("acquisitionservices");
const stm32cubemonitor_stlink_protocol_1 = __importDefault(require("@stm32/stm32cubemonitor-stlink-protocol"));
const events_1 = require("events");
const fp_1 = __importDefault(require("lodash/fp"));
const nanoSec = BigInt(1e9); // to manage the elapsed time in nanosecond
const sabAcqCountIndex = 0; // index of sab array in which is the number of acquisition is ongoing
const sabLockIndex = 1; // index of sab array in which is the state if the lock.
const locked = 1;
const unlocked = 0;
/**
 * This class aims at offering some generic services to interface with various probe drivers (as stlinkprotocol).
 * It inherits from EventEmitter class.
 */
class AcquisitionServices extends events_1.EventEmitter {
    /**
     * This static service aims at discovering the probes connected to the host.
     *
     * @param sab - SharedArrayBuffer
     * @param callback - this function is called back with the array of discovered probes
     * A single probe is represented in the array by an object holding a probeName string property,
     * a probeId string property , and an object property property.
     */
    static probesDiscovery(sab, callback) {
        let updated = false;
        if (Atomics.load(sab, sabAcqCountIndex) === 0) {
            AcquisitionServices.listProbe = stm32cubemonitor_stlink_protocol_1.default.probeDiscovery("", AcquisitionServices.connectionType);
            AcquisitionServices.lastUpdate = Date.now();
            updated = true;
        }
        callback(AcquisitionServices.listProbe, updated);
    }
    /**
     * This service aims at identifying a stLink probe by blinking its LED.
     *
     * @param sab - SharedArrayBuffer
     * @param probeId - the identifier of a probe
     * @returns the result of the blink command on the probe
     */
    static blinkLed(sab, probeId) {
        logger.debug("blinkLed on ", probeId);
        if (Atomics.load(sab, sabAcqCountIndex) !== 0) {
            return { status: false, err: "Acquisition ongoing" };
        }
        const res = stm32cubemonitor_stlink_protocol_1.default.blinkLed(probeId, AcquisitionServices.connectionType);
        if (res.result === 1 /* resultCode.STLINK_OK */) {
            return { status: true, err: "" };
        }
        const errLvl = res.result === 255 /* resultCode.STLINK_JTAG_CMD_NOT_SUPPORTED */ ? "warning" : "error";
        return { status: false, err: stm32cubemonitor_stlink_protocol_1.default.errorString(res.result), errLevel: errLvl };
    }
    /**
     * Set the connection type to be used between the driver and the probe.
     *
     * @param connectionType - The requested connection type, could be either "p2p" or "tcp"
     * @returns - The actual connection type that will be used in further commands
     */
    static initConnectionType(connectionType) {
        logger.debug("initConnectionType with ", connectionType);
        if (connectionType === "tcp") {
            AcquisitionServices.connectionType = stm32cubemonitor_stlink_protocol_1.default.isTcpAvailable() ? "tcp" : "p2p";
        }
        else {
            AcquisitionServices.connectionType = connectionType;
        }
        return AcquisitionServices.connectionType;
    }
    /**
     *
     * Initialize the logger with the provided level.
     * @param level - The level of logger, could be either off, fatal, error, warn, debug, info, trace.
     */
    static initLogger(level) {
        (0, stm32cubemonitor_logger_1.updateLevel)(level);
    }
    /**
     * Instantiate a new  Acquisitionservices class object for a dedicated probe and configure the future
     * communication link between the probe and the MCU target
     * @param probeId - the identifier of a probe for which a new AcquisitionServices object is instantiated.
     * This probeId is a public member of the object.
     * @param reconnectTime - the interval between 2 connections to the probe.
     * @param option - object holding 2 properties to configure the communication link between the probe
     * and the MCU target :
     *    protocol : protocol to be used by the probe to communicate with the target MCU
     *    frequency : the frequency to be used by the probe to communicate with the target MCU
     *    connectionType: the type of the connection to be used by the stlink to communicate with the probe,
     *    either p2p or tcp mode.
     * @param sab - SharedArrayBuffer
     */
    constructor(probeId, reconnectTime, option, sab) {
        super();
        this.probeId = probeId;
        this.reconnectTime = reconnectTime;
        this.option = option;
        this.sab = sab;
        /**
         * timeout set to relaunch connect() function every "reconnectTime" .
         */
        this.tout = null;
        logger.debug("call to new acquisitionservices ", probeId, reconnectTime, option);
        this.groups = {};
    }
    /**
     * This public service aims at actually connecting the probe to the MCU target.
     */
    connect() {
        let status;
        // Wait while is locked
        Atomics.wait(this.sab, sabLockIndex, locked);
        // Locked for the others
        Atomics.store(this.sab, sabLockIndex, locked);
        if (Atomics.load(this.sab, sabAcqCountIndex) === 0) {
            if (Date.now() - AcquisitionServices.lastUpdate > this.reconnectTime) {
                AcquisitionServices.lastUpdate = Date.now();
                status = this.connectToTarget(true);
            }
            else {
                status = this.connectToTarget(false);
            }
            if (status) {
                this.closeProbes();
            }
        }
        Atomics.store(this.sab, sabLockIndex, unlocked);
        Atomics.notify(this.sab, sabLockIndex, 1);
        if (this.tout != null) {
            clearTimeout(this.tout);
            this.tout = null;
        }
        this.tout = setTimeout(() => {
            this.connect();
        }, this.reconnectTime);
    }
    /**
     * This public service aims at closing the connection between the probe and the MCU target
     * @param cb - function called back when the connection is effectively closed
     */
    close() {
        logger.debug("close", this.probeId, this.option.connectionType);
        for (const groupname in this.groups) {
            if (this.groups.hasOwnProperty(groupname)) {
                this.stopAcquisition(groupname);
            }
        }
        if (this.tout != null) {
            clearTimeout(this.tout);
            this.tout = null;
        }
        this.closeProbes();
        logger.debug("Emit close" + this.probeId);
        this.emit("close");
    }
    /**
     * This service aims at sending a command to the connected probe. The list of possible commands are :
     *  read : to perform a single reading of memory target
     *  write : to write some values in memory target
     *  start: to start an acquisition loop at a configurable frequency
     *  stop : to stop an ongoing acquisition.
     * @param commandType - represents the command to send
     * @param data - object holding various configuration parameters for the command such as groupname,
     * variablelist, frequency
     * @param callback - function called back in case of error (parameter missing, ...)
     */
    sendCommand(commandType, data) {
        logger.debug("Arguments : commandType = ", commandType, ", data = ", data, ")");
        const myData = fp_1.default.cloneDeep(data);
        if (commandType) {
            const command = commandType.toLowerCase();
            switch (command) {
                case "read":
                    this.read(myData);
                    break;
                case "start":
                    this.start(myData);
                    break;
                case "stop":
                    this.stop(myData);
                    break;
                case "write":
                    this.write(myData);
                    break;
                default:
                    const resp = {
                        payload: "Unknown command",
                        topic: "error"
                    };
                    this.emit("error", resp);
                    break;
            }
        }
    }
    /**
     * This private service implements the read command.
     * @param myData - object holding various configuration parameters for the command such as groupname,
     * variablelist, frequency
     * @param callback - function called back in case of error (parameter missing, ...)
     */
    read(myData) {
        if (!this.connectToTarget(false)) {
            const resp = {
                payload: "Probe " + this.probeId + " not connected",
                topic: "error"
            };
            this.emit("error", resp);
            return;
        }
        this.readMemory(this, myData, BigInt(0), true);
        this.closeProbes();
    }
    /**
     * This private service implements the start command.
     * @param myData - object holding various configuration parameters for the command such as groupname,
     * variablelist, frequency
     * @param callback - function called back in case of error (parameter missing, ...)
     */
    start(myData) {
        if (!this.connectToTarget(false)) {
            const resp = {
                payload: "Probe " + this.probeId + " not connected",
                topic: "error"
            };
            this.emit("error", resp);
            return;
        }
        if (myData.hasOwnProperty("groupname") &&
            myData.hasOwnProperty("variablelist") &&
            myData.hasOwnProperty("frequency") &&
            myData.variablelist.length > 0) {
            // if acquisition already started on this group for this probe, don't start a new one
            if (this.groups[myData.groupname]) {
                return;
            }
            const variableslist = myData.variablelist.map((variable) => {
                const myvar = fp_1.default.cloneDeep(variable);
                myvar.values = [];
                return myvar;
            });
            const myGroup = {
                first: true,
                timerRead: {},
                variables: variableslist
            };
            this.groups[myData.groupname] = myGroup;
            Atomics.add(this.sab, sabAcqCountIndex, 1);
            this.startTimer = process.hrtime.bigint();
            let period = BigInt(0);
            if (myData.frequency > 0) {
                period = nanoSec / BigInt(myData.frequency);
                myData.timeToReach = this.startTimer + period;
            }
            myGroup.timerRead = setImmediate(() => {
                this.readMemory(this, myData, period, false);
            });
        }
        else {
            const resp = {
                payload: "Parameter missing",
                topic: "error"
            };
            this.emit("error", resp);
        }
    }
    /**
     * This private service implements the stop command.
     * @param myData - object holding various configuration parameters for the command such as groupname,
     * variablelist, frequency
     * @param callback - function called back in case of error (parameter missing, ...)
     */
    stop(myData) {
        this.stopAcquisition(myData.groupname);
        this.closeProbes();
    }
    /**
     * This private service implements the write command.
     * @param myData - object holding various configuration parameters for the command such as groupname,
     * variablelist, frequency
     * @param callback - function called back in case of error (parameter missing, ...)
     */
    write(myData) {
        // if acquisition already on going for this probe, don't re open it
        if (Object.entries(this.groups).length === 0) {
            if (!this.connectToTarget(false)) {
                const resp = {
                    payload: "Probe " + this.probeId + " not connected",
                    topic: "error"
                };
                this.emit("error", resp);
                return;
            }
        }
        logger.debug("call to stlinkprotocol.writeMemoryGroup ", this.probeId, myData, this.option.connectionType);
        const { result } = stm32cubemonitor_stlink_protocol_1.default.writeMemoryGroup(this.probeId, myData, this.option.connectionType);
        if (result === 1 /* resultCode.STLINK_OK */) {
            logger.debug("Successfull write on ", this.probeId, ", data = ", myData, ")");
        }
        else {
            const errorString = stm32cubemonitor_stlink_protocol_1.default.errorString(result);
            const error = "Unsuccessful write (" + errorString + ") on " + this.probeId + ", for data = " + JSON.stringify(myData);
            logger.warn(error);
            const resp = {
                payload: errorString,
                topic: "error"
                // TODO change error type according to error
            };
            this.emit("error", resp);
        }
        // if acquisition already on going for this probe, don't close it
        if (Object.entries(this.groups).length === 0) {
            this.closeProbes();
        }
    }
    /**
     * This private service aims at actually connecting the probe to the MCU target.
     * @param probeDiscovery - boolean to request for new probe discovery before open of the targeted probe.
     */
    connectToTarget(probeDiscovery) {
        logger.debug("call to stlinkprotocol.connectToTarget ", this.probeId, this.option.protocol, this.option.frequency, probeDiscovery, this.option.connectionType);
        const res = stm32cubemonitor_stlink_protocol_1.default.connectToTarget(this.probeId, this.option.protocol, this.option.frequency, probeDiscovery, this.option.connectionType);
        if (res.result === 1 /* resultCode.STLINK_OK */) {
            logger.debug("Emit open following the success of connect on target ", res.MCUName ? `(${res.MCUName})` : "");
            this.emit("open", res.MCUName ? `(${res.MCUName})` : "");
            return true;
        }
        else {
            const errorString = stm32cubemonitor_stlink_protocol_1.default.errorString(res.result);
            const resp = {
                payload: errorString,
                topic: "warning"
            };
            this.emit("error", resp);
            logger.debug("Emit close because failure of connect to target : ", errorString);
            this.emit("close", errorString);
            return false;
        }
    }
    /**
     * This private service manage the timer.
     * @param context - AcquisitionServices object context on which the reading is actually done.
     * @param data - parameters of the read command
     * @param callback - function called back on error
     */
    timeElapse(context, data, callback, period) {
        const timeNanoS = process.hrtime.bigint();
        if (timeNanoS < data.timeToReach) {
            context.groups[data.groupname].timerRead = setImmediate(() => {
                context.timeElapse(context, data, false, period);
            });
        }
        else {
            data.timeToReach = data.timeToReach + period;
            return context.readMemory(context, data, period, callback);
        }
    }
    /**
     * Internal service used to read values in MCU memory target at a given frequency
     * @param context - AcquisitionServices object context on which the reading is actually done.
     * @param data - parameters of the read command
     * @param oneshot - only one single reading of memory target is done if set to true
     * @param period - duration between two measurments
     */
    readMemory(context, data, period, oneshot) {
        logger.trace("Arguments : context.probeId = ", context.probeId, ", data = ", data, ", oneshot = ", oneshot);
        logger.debug("call to stlinkprotocol.readMemoryGroup ", context.probeId, data, this.option.connectionType);
        const res = stm32cubemonitor_stlink_protocol_1.default.readMemoryGroup(context.probeId, data, this.option.connectionType);
        if (res.result === 1 /* resultCode.STLINK_OK */ && res.data) {
            if (data.mode === "direct") {
                const resp = {
                    data: res.data,
                    first: oneshot || context.groups[data.groupname].first,
                    groupname: data.groupname
                };
                context.emit("data", resp);
            }
            else {
                const newDataFormat = [[]];
                for (let j = 0; j < res.data[0].length; j++) {
                    for (let i = 0; i < res.data.length; i++) {
                        newDataFormat[i] = [];
                        newDataFormat[i][0] = res.data[i][j];
                    }
                    const resp = {
                        data: newDataFormat,
                        first: oneshot || context.groups[data.groupname].first,
                        groupname: data.groupname
                    };
                    context.emit("data", resp);
                    context.groups[data.groupname].first = false;
                }
            }
            if (!oneshot) {
                context.groups[data.groupname].first = false;
                if (data.frequency === 0) {
                    context.groups[data.groupname].timerRead = setImmediate(() => {
                        context.readMemory(context, data, period);
                    });
                }
                else if (!context.option.acquFrequThreshold || data.frequency > context.option.acquFrequThreshold) {
                    context.groups[data.groupname].timerRead = setImmediate(() => {
                        context.timeElapse(context, data, false, period);
                    });
                }
                else {
                    let delay = 0;
                    const timeNanoS = process.hrtime.bigint();
                    if (data.timeToReach > timeNanoS) {
                        delay = Number(data.timeToReach - timeNanoS) / 1000000;
                    }
                    context.groups[data.groupname].timerRead = setTimeout(function () {
                        context.readMemory(context, data, period);
                    }, delay);
                    data.timeToReach = data.timeToReach + period;
                }
            }
        }
        else {
            /* res.result != STLINK_OK or no data*/
            context.stopAcquisition(data.groupname);
            let errorMessage;
            if (res.result === 1 /* resultCode.STLINK_OK */) {
                errorMessage = "No data received";
            }
            else {
                if (typeof res.result === "number") {
                    errorMessage = stm32cubemonitor_stlink_protocol_1.default.errorString(res.result)
                        ? stm32cubemonitor_stlink_protocol_1.default.errorString(res.result)
                        : "Unknown driver error";
                }
                else {
                    // error comes from snapshot
                    errorMessage = res.result;
                }
                const resp = {
                    payload: errorMessage,
                    topic: "error"
                };
                context.emit("error", resp);
            }
            // in case error is not specific to read operation on erroneous address
            if (res.result !== 244 /* resultCode.STLINK_JTAG_READ_ERROR */) {
                logger.warn("readMemory : emit close because of ", errorMessage);
                context.emit("close", errorMessage);
            }
            const resp = {
                payload: errorMessage,
                topic: "error"
            };
            context.emit("error", resp);
        }
    }
    /**
     * Private method used to stop an ongoing acquisition for a dedicated group
     * @param groupname - string representing the name of the group for which the acquisition must be stopped
     * @param callback - function called back if no acquisition is ongoing for the provided groupname
     */
    stopAcquisition(groupname) {
        const group = this.groups[groupname];
        if (group) {
            logger.debug("call to stlinkprotocol.stopSnapshot ", this.probeId);
            const res = stm32cubemonitor_stlink_protocol_1.default.stopSnapshot(this.probeId);
            if (res !== 1 /* resultCode.STLINK_OK */) {
                const resp = {
                    payload: res,
                    topic: "warning"
                };
                this.emit("error", resp);
            }
            if (group.timerRead.hasOwnProperty("_onImmediate")) {
                clearImmediate(group.timerRead);
            }
            else {
                clearTimeout(group.timerRead);
            }
            this.emit("stop");
            delete this.groups[groupname];
            Atomics.sub(this.sab, sabAcqCountIndex, 1);
        }
        else {
            const resp = {
                payload: "Group " + groupname + " is not defined or stopped",
                topic: "warning"
            };
            this.emit("error", resp);
        }
    }
    /**
     * Private method used to close openned probes
     */
    closeProbes() {
        logger.debug("call to stlinkprotocol.closeProbes ", this.probeId, this.option.connectionType);
        const res = stm32cubemonitor_stlink_protocol_1.default.closeProbes(this.probeId, this.option.connectionType);
        if (res.result !== 1 /* resultCode.STLINK_OK */) {
            logger.warn("Could not close probes because of error: " + res.result);
            const resp = {
                payload: stm32cubemonitor_stlink_protocol_1.default.errorString(res.result),
                topic: "warning"
            };
            this.emit("error", resp);
        }
    }
}
exports.AcquisitionServices = AcquisitionServices;
/**
 * Date of the last discovery of connected probes.
 */
AcquisitionServices.lastUpdate = Date.now();
