/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const logger = require("@stm32/stm32cubemonitor-logger").getLogger("stlinkprotocol");
let ref = require("@breush/ref-napi");
const koffi = require("koffi");
let StructType = require("ref-struct-di")(ref);
let ArrayType = require("ref-array-di")(ref);
let path = require("path");
let fs = require("fs");
let tmp = require("tmp");
let os = require("os");

// definition of constants used in DLL interface
let defs = {
  DRV_OK: 0x01,
  DRV_ERR: 0x04,
  DRV_NO_DEVICE: 0x82,
  DRV_INVALID_HA: 0x81,
  DRV_INVALID_SRB: 0xe0,
  DRV_FAILED_INIT: 0xe4,
  DRV_SKIP_STEP: 0xf1,

  DRV_MEMORY_PROBLEM: 0x1000,
  DRV_TIMEOUT: 0x1001,
  DRV_BAD_PARAMETER: 0x1002,
  DRV_OPEN_ERR: 0x1003,
  DRV_TRANSFER_ERR: 0x1004,
  DRV_NO_SUCH_INFO: 0x1050,
  DRV_BAD_BUFFER_SIZE: 0x1051,
  DRV_TRUNCATED_DATA: 0x1052,
  DRV_CMD_NOT_AVAILABLE: 0x1053,
  DRV_DEVICE_NOT_SUPPORTED: 0x1054,
  DRV_PERMISSION_ERR: 0x1055,

  DRV_FAILED: 0x2000,
  DRV_TCP_ERROR: 0x2001, // To be refined later
  DRV_TCP_CANT_CONNECT: 0x2002, // Error when TCP client is not able to connect to the server
  DRV_TCP_CLOSE_ERROR: 0x2003, // Error to close the TCP socket (remaining STLink opened ...)

  DRV_SCSI_ERROR: 0x5000,
  DRV_WIN32_ERROR: 0x10000,

  DRV_TCP_ERROR_SIZE: 4,

  DRV_DEFAULT_TIMEOUT: 5000, // ms

  DRV_REQUEST_WRITE: 0, //Through EP2OUT
  DRV_REQUEST_READ_EP3: 3, //Through EP3IN (SWV)
  DRV_REQUEST_READ_EP1: 1, // Through EP1IN

  DRV_SERIAL_NUM_STR_MAX_LEN: 32,

  DRV_DEFAULT_SENSE_LEN: 14,

  DRV_CLOSE_TCP_AUTO: 0, //constant defining close type in STLink_CloseDeviceTcp() closeTcp parameter
  DRV_CLOSE_ONLY_DEVICE: 1, // Close the STLink device but do not close the TCP socket : need to call later STLink_CloseDeviceTcp(CLOSE_ONLY_TCP)
  DRV_CLOSE_ONLY_TCP: 2, // Close the TCP socket (and potentially stlink-server) if no remaining opened STLink device

  DRV_DEFAULT_IF: 0, // Debug Interface
  DRV_STLINK_TCP: 2,

  DRV_OPEN_STLINK_EXCLUSIVE_ACCESS: 1,
  DRV_OPEN_STLINK_SHARED_ACCESS: 0
};

// definition of types used on the DLL interface
let TDeviceInfo2 = StructType(
  {
    StLinkUsbId: ref.types.int32, // DWORD - Tcp server Device cookie in little endian format, to use in STLINK_TCP_CMD_OPEN_DEVICE
    EnumUniqueId: ArrayType(ref.types.char, defs.DRV_SERIAL_NUM_STR_MAX_LEN), // Unique instance ID from system enumeration (equal to device Serial number in some cases, but not always ...)
    VendorId: ref.types.uint16, // Vendor ID from USB device descriptor (system enumeration)
    ProductId: ref.types.uint16, // Product ID from USB device descriptor (system enumeration)
    DeviceUsed: ref.types.byte // On windows, equal to 1 if device interface was already opened from externally when enumerating or trying to open
  },
  { packed: true }
);
// structure declaration for Koffi mapping, must be aligned with TDeviceInfo2
let KTDeviceInfo2 = koffi.struct("KTDeviceInfo2", {
  StLinkUsbId: "uint32_le", // DWORD - Tcp server Device cookie in little endian format, to use in STLINK_TCP_CMD_OPEN_DEVICE
  EnumUniqueId: koffi.array("char", defs.DRV_SERIAL_NUM_STR_MAX_LEN, "Array"), // Unique instance ID from system enumeration (equal to device Serial number in some cases, but not always ...)
  VendorId: "uint16_t", // Vendor ID from USB device descriptor (system enumeration)
  ProductId: "uint16_t", // Product ID from USB device descriptor (system enumeration)
  DeviceUsed: "uint8_t" // On windows, equal to 1 if device interface was already opened from externally when enumerating or trying to open
});

// Structure used to send request to Stlink
let TDeviceRequest = StructType(
  {
    CDBLength: ref.types.byte, // Command header length (CDB on mass storage) - used only on ST-Link/V1
    CDBByte: ArrayType(ref.types.byte, 16), // Command header (CDB on mass storage)
    InputRequest: ref.types.byte, // 0 (output) or input on ST-Link/V1, REQUEST_WRITE, REQUEST_READ_EP1 or REQUEST_READ_EP3 on ST-Link/V2
    Buffer: ref.refType(ref.types.void), // Data for data stage (if any)
    BufferLength: ref.types.int32, // DWORD - Size of data stage (0 if no data stage)
    SenseLength: ref.types.byte, // To be initialized: DEFAULT_SENSE_LEN - used only on ST-Link/V1
    Sense: ArrayType(ref.types.byte, 16) // Used only on ST-Link/V1
  },
  { packed: true }
);
// Request structure declaration for Koffi mapping
let KBuffer = koffi.opaque("KBuffer");
let KTDeviceRequest = koffi.struct("KTDeviceRequest", {
  CDBLength: "uint8_t", // Command header length (CDB on mass storage) - used only on ST-Link/V1
  CDBByte: koffi.array("uint8_t", 16), // Command header (CDB on mass storage)
  InputRequest: "uint8_t", // 0 (output) or input on ST-Link/V1, REQUEST_WRITE, REQUEST_READ_EP1 or REQUEST_READ_EP3 on ST-Link/V2
  KBuffer: "KBuffer *", // Data for data stage (if any)
  BufferLength: "uint32_t", // DWORD - Size of data stage (0 if no data stage)
  SenseLength: "uint8_t", // To be initialized: DEFAULT_SENSE_LEN - used only on ST-Link/V1
  Sense: koffi.array("uint8_t", 16) // Used only on ST-Link/V1
});

// Declare Handle for the open function
const Handle = koffi.opaque("Handle");

// DLL Loading + DLL interface definition
let platform = process.platform;
let stLinkUsbDriver = null;
let stLinkDrv, stLinkDrvFfiMapping;

// remove the TMP env variable in the worker. (issue with linux)
process.env.TMP = "";
// create tmp folder
let tmpDir = path.join(os.tmpdir(), "stm32cubemon-tmp");
if (!fs.existsSync(tmpDir)) {
  fs.mkdirSync(tmpDir);
}
let tmpLibPath;
if (platform === "win32") {
  tmpLibPath = tmp.tmpNameSync({ dir: "stm32cubemon-tmp", prefix: "stm32cubemon", postfix: ".dll" });
  // Build dll path. In case of asar pack, use the "unpacked" path for the dll.
  stLinkUsbDriver = path.join(__dirname, "STLinkUSBDriver.dll").replace("app.asar", "app.asar.unpacked");
} else if (platform === "linux") {
  tmpLibPath = tmp.tmpNameSync({ dir: "stm32cubemon-tmp", prefix: "stm32cubemon", postfix: ".so" });
  // Build dll path. In case of asar pack, use the "unpacked" path for the dll.
  stLinkUsbDriver = path.join(__dirname, "libSTLinkUSBDriver.so").replace("app.asar", "app.asar.unpacked");
} else if (platform === "darwin") {
  tmpLibPath = tmp.tmpNameSync({ dir: "stm32cubemon-tmp", prefix: "stm32cubemon", postfix: ".dylib" });
  // Build dll path. In case of asar pack, use the "unpacked" path for the dll.
  stLinkUsbDriver = path.join(__dirname, "libSTLinkUSBDriver.dylib").replace("app.asar", "app.asar.unpacked");
} else {
  throw new Error("unsupported plateform :" + platform + " for stLinkUsbDriver");
}

// copy the lib in the tmp folder
fs.copyFileSync(stLinkUsbDriver, tmpLibPath);
logger.debug("St Link dll path :", tmpLibPath);

// DLL function mapping
if (platform === "win32") {
  STLinkLib = koffi.load(tmpLibPath);
  stLinkDrvFfiMapping = {
    STLink_Reenumerate: STLinkLib.func("int STLink_Reenumerate( int IfId, bool bClearList)"),
    STLink_ReenumerateTcp: STLinkLib.func(
      "int STLink_ReenumerateTcp (int IfId, bool bClearList, string pConnectParams, string pServerCmdLineParams)"
    ),
    STLink_GetNbDevices: STLinkLib.func("int STLink_GetNbDevices (int IfId)"),
    STLink_GetDeviceInfo2: STLinkLib.func(
      "int STLink_GetDeviceInfo2(int IfId, int8_t DevIdxInList,_Out_ KTDeviceInfo2 *pInfo, long InfoSize)"
    ),
    STLink_OpenDevice: STLinkLib.func(
      "int STLink_OpenDevice( int IfId, uint8_t DevIdxInList, uint8_t bExclusiveAccess,_Out_ Handle **handle)"
    ),
    STLink_OpenDeviceTcp: STLinkLib.func(
      "int STLink_OpenDeviceTcp( uint32_le StLinkUsbId,uint32_t DevInfoUsbId, uint8_t bExclusiveAccess)"
    ),
    STLink_CloseDevice: STLinkLib.func("int STLink_CloseDevice( Handle *h)"),
    STLink_CloseDeviceTcp: STLinkLib.func("int STLink_CloseDeviceTcp( uint32_le StLinkUsbId, uint8_t closeTcp)"),
    STLink_SendCommand: STLinkLib.func(
      "int STLink_SendCommand(Handle *h, _Inout_ KTDeviceRequest *req, long DwTimeOut)"
    ),
    STLink_SendCommandTcp: STLinkLib.func(
      "int STLink_SendCommandTcp(uint32_le StLinkUsbId, _Inout_ KTDeviceRequest *req, long DwTimeOut)"
    )
  };
} else if (platform === "linux" || platform === "darwin") {
  STLinkLib = koffi.load(tmpLibPath);
  stLinkDrvFfiMapping = {
    STLink_Reenumerate: STLinkLib.func("int _Z18STLink_Reenumerate20TEnumStlinkInterfaceh (int IfId, bool bClearList)"),
    STLink_ReenumerateTcp: STLinkLib.func(
      "int _Z21STLink_ReenumerateTcp20TEnumStlinkInterfacehPcS0_ (int IfId, bool bClearList, string pConnectParams, string pServerCmdLineParams)"
    ),
    STLink_GetNbDevices: STLinkLib.func("int _Z19STLink_GetNbDevices20TEnumStlinkInterface(int IfId)"),
    STLink_GetDeviceInfo2: STLinkLib.func(
      "int _Z21STLink_GetDeviceInfo220TEnumStlinkInterfacehP12TDeviceInfo2j(int IfId, int8_t DevIdxInList,_Out_ KTDeviceInfo2 *pInfo, long InfoSize)"
    ),
    STLink_OpenDevice: STLinkLib.func(
      "int _Z17STLink_OpenDevice20TEnumStlinkInterfacehhPPv( int IfId, uint8_t DevIdxInList, uint8_t bExclusiveAccess,_Out_ Handle **handle)"
    ),
    STLink_OpenDeviceTcp: STLinkLib.func(
      "int _Z20STLink_OpenDeviceTcp20TEnumStlinkInterfacejh(uint32_le StLinkUsbId, uint32_t DevInfoUsbId, uint8_t bExclusiveAccess)"
    ),
    STLink_CloseDevice: STLinkLib.func("int _Z18STLink_CloseDevicePv( Handle *h)"),
    STLink_CloseDeviceTcp: STLinkLib.func("int _Z21STLink_CloseDeviceTcpjh( int StLinkUsbId, uint8_t closeTcp)"),
    STLink_SendCommand: STLinkLib.func(
      "int _Z18STLink_SendCommandPvP14TDeviceRequestj(Handle *h, _Inout_ KTDeviceRequest *req, long DwTimeOut)"
    ),
    STLink_SendCommandTcp: STLinkLib.func(
      "int _Z21STLink_SendCommandTcpjP14TDeviceRequestj(int a, _Inout_ KTDeviceRequest *req, long DwTimeOut)"
    )
  };
} else {
  throw new Error("unsupported plateform :" + platform + " for stLinkUsbDriver");
}

stLinkDrv = (function () {
  let map = {};
  for (let [key, value] of Object.entries(stLinkDrvFfiMapping)) {
    // eslint-disable-next-line require-jsdoc
    let callDriverAPI = (...args) => {
      logger.debug(key, args);
      return value.apply(this, args);
    };
    //property writable set to true for unitary test purpose
    map = Object.defineProperty(map, key, { value: callDriverAPI, writable: true });
  }
  return map;
})();

// module exports
module.exports.stLinkDrv = stLinkDrv;
module.exports.TDeviceInfo2 = TDeviceInfo2;
module.exports.TDeviceRequest = TDeviceRequest;
module.exports.defs = defs;
